#! /usr/bin/env python

#  This Source Code Form is subject to the terms of the Mozilla Public
#  License, v. 2.0. If a copy of the MPL was not distributed with this
#  file, You can obtain one at http://mozilla.org/MPL/2.0/.

import argparse

from hydralink import HydraLink


def main() -> None:

    parser = argparse.ArgumentParser(
                        prog='hydralink_config',
                        description='Configures the dissecto HydraLink MAC and PHY')
    parser.add_argument('--gui', action='store_true')
    parser.add_argument('-g', '--gigabit', action='store_true')
    parser.add_argument('-m', '--master',  action='store_true')
    parser.add_argument('-d', '--device', type=int)
    parser.add_argument('-p', '--promiscuous', type=bool)
    parser.add_argument('--dump', type=str, help="Dump all MAC registers to file")
    args = parser.parse_args()

    if args.gui:
        import hydralink.gui
        return hydralink.gui.main()

    hl = HydraLink(args.device)

    if args.dump is not None:
        with open(args.dump, "w") as of:
            for i in range(0, 0x1000, 4):
                print("%03x: %08x" % (i, hl.mac[i]), file=of)
        return

    hl.setup(master=args.master,
             speed=(1000 if args.gigabit else 100),
             promiscuous=args.promiscuous)


if __name__ == '__main__':
    main()



#  This Source Code Form is subject to the terms of the Mozilla Public
#  License, v. 2.0. If a copy of the MPL was not distributed with this
#  file, You can obtain one at http://mozilla.org/MPL/2.0/.

import ctypes
import struct
import time

from typing import Optional, cast
from hydralink.windows_apis import CloseHandle, CreateFile, DeviceIoControl

from hydralink.lan7801 import LAN7801_LL


class LAN7801_Win(LAN7801_LL):
    class DriverError(Exception):
        def __init__(self, errno: int):
            self.errno = errno
            Exception.__init__(self, f"LAN7801 DeviceIoControl failed with error {errno}")

    def __init__(self, interface_idx: Optional[int] = None) -> None:
        self.handle = CreateFile(b"\\\\.\\LAN7800_IOCTL", 0xc0000000, 3, 0, 3, 0x80, 0)
        err = ctypes.GetLastError()
        if err == 2:
            raise FileNotFoundError("Device not found!")
        elif err:
            raise ctypes.WinError()

        self._buffer = bytearray(0x1000)
        self._buffer_c = (ctypes.c_char * 0x1000).from_buffer(self._buffer)
        self.iidx = 0
        self.oidx = 0x10002
        if interface_idx is not None:
            self.iidx = interface_idx
            self.oidx = 0x20001

    def __del__(self) -> None:
        CloseHandle(self.handle)

    def _xfer(self, buffer: bytes, timeout: float = 5) -> bytes:
        if len(buffer) > len(self._buffer):
            raise ValueError("Too much data to transfer")
        self._buffer[:len(buffer)] = buffer

        outsize = ctypes.c_uint32(0)
        deadline = time.monotonic() + timeout
        err = 4

        while err != 0:
            if not DeviceIoControl(
                    self.handle, 0x122400,
                    self._buffer_c, len(buffer),
                    self._buffer_c, len(self._buffer), ctypes.byref(outsize),
                    0):
                raise ctypes.WinError()

            if outsize.value > len(self._buffer):
                raise OverflowError(f"LAN7801 response too large ({outsize.value} > {len(self._buffer)})")
            if outsize.value < 16:
                raise IOError(f"LAN7801 response too small ({outsize.value} < 16))")

            err = struct.unpack("<I", self._buffer[8:12])[0]
            if err == 4:
                if time.monotonic() >= deadline:
                    raise TimeoutError("LAN7801 is busy, please retry later or reconnect the device")
            elif err != 0:
                raise LAN7801_Win.DriverError(err)

        return bytes(self._buffer[16:outsize.value])

    def write_reg(self, address: int, value: int) -> None:
        if value != value & 0xffffffff:
            raise ValueError("Value must be an unsigned 32-bit integer")
        if address != address & 0xfff:
            raise ValueError("Address must be an unsigned 12-bit integer")

        resbuf = self._xfer(struct.pack("<IIIII", self.oidx, 9, self.iidx, address, value))

        if len(resbuf) != 1:
            raise IOError(f"Unexpected response length ({len(resbuf)} != 1) to WRITE_REGISTER")

    def read_reg(self, address: int) -> int:
        if address != address & 0xfff:
            raise ValueError("Address must be an unsigned 12-bit integer")

        resbuf = self._xfer(struct.pack("<IIII", self.oidx, 8, self.iidx, address))

        if len(resbuf) != 4:
            raise IOError(f"Unexpected response length ({len(resbuf)} != 4) to READ_REGISTER")

        res = cast(int, struct.unpack("<I", resbuf)[0])
        return res

    def get_adapter_registry_key(self) -> str:
        resbuf = self._xfer(struct.pack("<III", self.oidx, 0x12, self.iidx))
        s = resbuf.decode('UTF-16')
        assert s[-1] == '\x00'
        return s[:-1]



#  This Source Code Form is subject to the terms of the Mozilla Public
#  License, v. 2.0. If a copy of the MPL was not distributed with this
#  file, You can obtain one at http://mozilla.org/MPL/2.0/.

import struct
import sys

from typing import Union, Optional

from hydralink.lan7801 import LAN7801, LAN7801_LL
from hydralink.bcm89881 import BCM89881


def is_windows() -> bool:
    return sys.platform in ['win32', 'cygwin', 'msys']


def get_lan7801(ll: Union[None, int, LAN7801_LL] = None) -> LAN7801:
    if isinstance(ll, LAN7801_LL):
        return LAN7801(ll)

    if is_windows():
        from hydralink.lan7801_win import LAN7801_Win
        return LAN7801(LAN7801_Win(ll))
    else:
        from hydralink.lan7801_libusb import LAN7801_LibUSB
        return LAN7801(LAN7801_LibUSB(ll))


class HydraLink:
    def __init__(self, ll: Union[None, int, LAN7801_LL] = None) -> None:
        self.mac = get_lan7801(ll)

        # Read MAC register
        identifier = self.mac[0]
        if (identifier >> 16) != 0x7801:
            raise IOError(f"Wrong MAC identifier: 0x{identifier:x}")

        # Access clause 45 registers
        self.phy = BCM89881(self.mac, 0)
        identifier = self.phy[1, 2]
        if identifier != 0xae02:
            raise IOError(f"Wrong PHY identifier: 0x{identifier:x}")

    def setup(self,
              master: Optional[bool] = None,
              speed: Optional[int] = None,
              mac_addr: Optional[str] = None,
              promiscuous: Optional[bool] = None
              ) -> None:
        mac = self.mac
        phy = self.phy

        phy.reset(True)

        # Enable clocks
        mac[0x010] |= 0x02000000
        # MAC-PHY RGMII clock delay setup
        mac[0x128] = 0x00000002
        phy[1, 0xa010] = 0x0001
        phy[1, 0xa015] = 0x0000
        # PHY LEDs setup
        phy[1, 0xa027] = 0x0f15
        phy[1, 0x931d] = 0x0010
        phy[1, 0x931e] = 0x0063

        if promiscuous is not None:
            if promiscuous:
                mac[0x0b0] = 0x1f80
                print("Enabled promiscuous mode")
            else:
                mac[0x0b0] = 0x1c8a
                print("Disabled promiscuous mode")

        if mac_addr is not None:
            mac_addr_bytes = b''
            for b in mac_addr.split(':'):
                bb = bytes.fromhex(b)
                if len(bb) != 1:
                    raise ValueError("Malformed MAC address")
                mac_addr_bytes += bb
            if len(mac_addr_bytes) != 6:
                raise ValueError("Malformed MAC address")
            hi, lo = struct.unpack(">HI", mac_addr_bytes)
            mac[0x118] = hi
            mac[0x11c] = lo

        if speed is not None:
            # Disable Automatic Speed Detection
            mac.set_ads(False)
            if speed == 1000:
                mac.set_speed(2)
                phy.set_speed(1000)
                print("Set hydralink speed to 1 Gb/s")
            elif speed == 100:
                mac.set_speed(1)
                phy.set_speed(100)
                print("Set hydralink speed to 100 Mb/s")
            else:
                raise ValueError("Speed should be either 100 or 1000")

        if master is not None:
            phy.set_master(master)
            print("Set hydralink to operate as %s" % ("master" if master else "slave"))

        # Deassert reset, resume operation
        phy.reset(False)



#  This Source Code Form is subject to the terms of the Mozilla Public
#  License, v. 2.0. If a copy of the MPL was not distributed with this
#  file, You can obtain one at http://mozilla.org/MPL/2.0/.

from typing import Optional
from hydralink.lan7801 import LAN7801


class BCM89881:
    def __init__(self, mac: LAN7801, phy_addr: int):
        if phy_addr & 0x1f != phy_addr:
            raise ValueError('PHY address must be a 5-bit unsigned integer')
        self.mac = mac
        self.phy_addr = phy_addr

    def __getitem__(self, key: tuple[int, int]) -> int:
        return self.mac.read_mdio_reg_c45(self.phy_addr, *key)

    def __setitem__(self, key: tuple[int, int], value: int) -> None:
        self.mac.write_mdio_reg_c45(self.phy_addr, *key, value)

    def edit_register(self, devaddr: int, miiaddr: int, set: int, clr: int) -> int:
        if set & 0xffff != set:
            raise ValueError("set mask bust be a 16-bit unsigned integer")
        if clr & 0xffff != clr:
            raise ValueError("set mask bust be a 16-bit unsigned integer")
        if clr & set:
            raise ValueError("set and clr masks are conflicting")
        reg = self.mac.read_mdio_reg_c45(self.phy_addr, devaddr, miiaddr)
        old_reg = reg
        reg &= (~clr) & 0xffff
        reg |= set
        if old_reg != reg:
            self.mac.write_mdio_reg_c45(self.phy_addr, devaddr, miiaddr, reg)
        return reg

    def reset(self, rst: bool) -> None:
        if rst:
            self.edit_register(1, 0x0000, 0x8000, 0x0000)
        else:
            self.edit_register(1, 0x0000, 0x0000, 0x8000)

    def set_speed(self, speed: int) -> None:
        if speed == 1000:
            self.edit_register(1, 0x0000, 0x2000, 0x0040)
        elif speed == 100:
            self.edit_register(1, 0x0000, 0x0040, 0x2000)
        else:
            raise ValueError("Speed must be 100 or 1000")

    def set_master(self, master: bool) -> None:
        if master:
            self.edit_register(1, 0x0834, 0x4000, 0x0000)
        else:
            self.edit_register(1, 0x0834, 0x0000, 0x4000)

    def get_speed(self) -> Optional[int]:
        s = self[1, 0]
        if s & 0x2000:
            return 1000
        elif s & 0x0040:
            return 100
        else:
            return None

    def get_master(self) -> bool:
        return (self[1, 0x0834] & 0x4000) == 0x4000



#  This Source Code Form is subject to the terms of the Mozilla Public
#  License, v. 2.0. If a copy of the MPL was not distributed with this
#  file, You can obtain one at http://mozilla.org/MPL/2.0/.

class LAN7801_LL:

    def write_reg(self, address: int, value: int) -> None:
        raise NotImplementedError()

    def read_reg(self, address: int) -> int:
        raise NotImplementedError()


class LAN7801:
    def __init__(self, dev: LAN7801_LL) -> None:
        self.dev = dev
        self.write_reg = dev.write_reg
        self.read_reg = dev.read_reg

    def read_mdio_reg(self, phy_addr: int, miirinda: int) -> int:
        if phy_addr != phy_addr & 0x1f:
            raise ValueError("PHY address must be a 5-bit unsigned integer")
        if miirinda != miirinda & 0x1f:
            raise ValueError("MII Register Index must be a 5-bit unsigned integer")
        miibzy = 1
        while miibzy:
            mii_access = self.read_reg(0x120)
            miibzy = mii_access & 1
        self.write_reg(0x120, (phy_addr << 11) | (miirinda << 6) | 1)
        while miibzy:
            mii_access = self.read_reg(0x120)
            miibzy = mii_access & 1
        return self.read_reg(0x124)

    def write_mdio_reg(self, phy_addr: int, miirinda: int, data: int) -> None:
        if data != data & 0xffff:
            raise ValueError("MII data must be a 16-bit unsigned integer")
        if phy_addr != phy_addr & 0x1f:
            raise ValueError("PHY address must be a 5-bit unsigned integer")
        if miirinda != miirinda & 0x1f:
            raise ValueError("MII Register Index must be a 5-bit unsigned integer")
        miibzy = 1
        while miibzy:
            mii_access = self.read_reg(0x120)
            miibzy = mii_access & 1
        self.write_reg(0x124, data)
        self.write_reg(0x120, (phy_addr << 11) | (miirinda << 6) | 3)
        while miibzy:
            mii_access = self.read_reg(0x120)
            miibzy = mii_access & 1

    def read_mdio_reg_c45(self, phy_addr: int, devad: int, miirinda: int) -> int:
        if phy_addr != phy_addr & 0x1f:
            raise ValueError("PHY address must be a 5-bit unsigned integer")
        if devad != devad & 0x1f:
            raise ValueError("DEVAD must be a 5-bit unsigned integer")
        if miirinda != miirinda & 0xffff:
            raise ValueError("MII Register Index must be a 5-bit unsigned integer")
        self.write_mdio_reg(phy_addr, 0xd, 0x0000 | devad)
        self.write_mdio_reg(phy_addr, 0xe, miirinda)
        self.write_mdio_reg(phy_addr, 0xd, 0x4000 | devad)
        return self.read_mdio_reg(phy_addr, 0xe)

    def write_mdio_reg_c45(self, phy_addr: int, devad: int, miirinda: int, data: int) -> None:
        if data != data & 0xffff:
            raise ValueError("MII data must be a 16-bit unsigned integer")
        if phy_addr != phy_addr & 0x1f:
            raise ValueError("PHY address must be a 5-bit unsigned integer")
        if devad != devad & 0x1f:
            raise ValueError("DEVAD must be a 5-bit unsigned integer")
        if miirinda != miirinda & 0xffff:
            raise ValueError("MII Register Index must be a 5-bit unsigned integer")
        self.write_mdio_reg(phy_addr, 0xd, 0x0000 | devad)
        self.write_mdio_reg(phy_addr, 0xe, miirinda)
        self.write_mdio_reg(phy_addr, 0xd, 0x4000 | devad)
        self.write_mdio_reg(phy_addr, 0xe, data)

    def __getitem__(self, key: int) -> int:
        if isinstance(key, int):
            return self.read_reg(key)
        else:
            raise ValueError("Unsupported key type")

    def __setitem__(self, key: int, value: int) -> None:
        if isinstance(key, int):
            return self.write_reg(key, value)
        else:
            raise ValueError("Unsupported key type")

    def set_ads(self, ads_enabled: bool) -> None:
        p = self[0x100]
        p &= ~0x0800
        if ads_enabled:
            p |= 0x0800
        self[0x100] = p

    def set_speed(self, speed: int) -> None:
        if speed & 3 != speed:
            raise ValueError("MAC Configuration must be a 2-bit unsigned integer")
        old_reg = self[0x100]
        p = old_reg
        p &= ~0x0006
        p |= (speed & 3) << 1
        if old_reg != p:
            self[0x100] = p

    def get_speed(self) -> int:
        reg = self[0x100]
        return (reg & 0x0006) >> 1



#  This Source Code Form is subject to the terms of the Mozilla Public
#  License, v. 2.0. If a copy of the MPL was not distributed with this
#  file, You can obtain one at http://mozilla.org/MPL/2.0/.

import usb.core
import struct

from typing import Union, cast

from hydralink.lan7801 import LAN7801_LL


class LAN7801_LibUSB(LAN7801_LL):
    def __init__(self, d: Union[None, int, usb.core.Device] = None) -> None:
        if isinstance(d, usb.core.Device):
            dev = d
        elif isinstance(d, int):
            dev = list(usb.core.find(find_all=True, idVendor=0x0424, idProduct=0x7801))[d]
        else:
            dev = usb.core.find(idVendor=0x0424, idProduct=0x7801)

        if dev is None:
            raise FileNotFoundError("Device not found!")
        self.dev = dev

    def write_reg(self, address: int, value: int) -> None:
        if value != value & 0xffffffff:
            raise ValueError("Value must be an unsigned 32-bit integer")
        if address != address & 0xfff:
            raise ValueError("Address must be an unsigned 12-bit integer")
        msg = struct.pack("<I", value)
        ret = self.dev.ctrl_transfer(0x40, 0xa0, 0, address, msg)
        if ret != 4:
            raise IOError("Written bytes is not 4")

    def read_reg(self, address: int) -> int:
        if address != address & 0xfff:
            raise ValueError("Address must be an unsigned 12-bit integer")
        ret = self.dev.ctrl_transfer(0xc0, 0xa1, 0, address, 4)
        if len(ret) != 4:
            raise IOError("Received response is not 4 bytes")
        val, = struct.unpack("<I", bytes(ret))
        return cast(int, val)



#  This Source Code Form is subject to the terms of the Mozilla Public
#  License, v. 2.0. If a copy of the MPL was not distributed with this
#  file, You can obtain one at http://mozilla.org/MPL/2.0/.

from tkinter import ttk, messagebox
import tkinter as tk
from typing import Any, Dict, List, NamedTuple, Optional
import hydralink.hydralink
from hydralink.hydralink import HydraLink


class FoundUsbDevice(NamedTuple):
    vid: int
    pid: int
    serialnum: str
    path: str
    software_key: str


if hydralink.hydralink.is_windows():
    from hydralink.windows_apis import list_usb_devices
    from hydralink.lan7801_win import LAN7801_Win

    def get_hydralinks() -> Dict[str, Any]:
        return {t.serialnum: t for t in list_usb_devices() if t.vid == 0x0424 and t.pid == 0x7801}

    def hydralink_by_serial(serial: str) -> Optional[HydraLink]:
        devices = get_hydralinks()
        if serial not in devices:
            return None
        key = devices[serial].software_key
        exceptions: List[Exception] = []
        for i in range(80):
            mac = LAN7801_Win(i)
            try:
                mac_key = mac.get_adapter_registry_key()
            except Exception as ex:
                exceptions.append(ex)
            else:
                print(key, '==', mac_key)
                if key == mac_key:
                    return HydraLink(mac)
            del mac

        if len(exceptions) == 0:
            # This should not really happen
            return None
        # Raise the exception most likely to have caused the search to fail
        for x in exceptions:
            if not (isinstance(x, WindowsError) and hasattr(x, 'winerror') and x.winerror == 31):
                raise x
        raise exceptions[0]
else:
    import usb.core
    from hydralink.lan7801_libusb import LAN7801_LibUSB

    def get_hydralinks() -> Dict[str, Any]:
        return {"%d.%d" % (dev.bus, dev.address) if dev.serial_number is None else dev.serial_number: dev
                for dev in usb.core.find(find_all=True, idVendor=0x0424, idProduct=0x7801)}

    def hydralink_by_serial(serial: str) -> Optional[HydraLink]:
        devices = get_hydralinks()
        if serial in devices:
            dev: usb.core.Device = devices[serial]
            return HydraLink(LAN7801_LibUSB(dev))
        else:
            return None


class Gui:
    def __init__(self) -> None:
        self.root = tk.Tk()
        self.root.title('HydraLink configuration')
        self.frm = ttk.Frame(self.root, padding=10)
        self.frm.grid()

        self.hl: Optional[HydraLink] = None

        self.speed_var = tk.IntVar()
        self.master_var = tk.IntVar()
        self.promiscuous_var = tk.IntVar()

        ttk.Label(self.frm, text="Device Selection:").grid(column=0, row=1)
        self.device_var = tk.StringVar()
        self.combobox = ttk.Combobox(self.frm, textvariable=self.device_var,
                                     state='readonly',
                                     postcommand=self.dropdown_opened)
        self.combobox.grid(column=1, row=1)
        self.combobox.bind('<<ComboboxSelected>>', self.change_device)

        self.button_gigabit = ttk.Radiobutton(self.frm, text='1 Gigabit/s', variable=self.speed_var,
                                              value=1000, command=self.update_settings,
                                              state='disabled')
        self.button_gigabit.grid(column=0, row=2)

        self.button_100mbit = ttk.Radiobutton(self.frm, text='100 Megabit/s', variable=self.speed_var,
                                              value=100, command=self.update_settings,
                                              state='disabled')
        self.button_100mbit.grid(column=1, row=2)

        self.button_master = ttk.Radiobutton(self.frm, text='Master', variable=self.master_var,
                                             value=1, command=self.update_settings,
                                             state='disabled')
        self.button_master.grid(column=0, row=3)

        self.button_slave = ttk.Radiobutton(self.frm, text='Slave', variable=self.master_var,
                                            value=0, command=self.update_settings,
                                            state='disabled')
        self.button_slave.grid(column=1, row=3)

        self.button_promiscuous = ttk.Checkbutton(self.frm, text='Promiscuous Mode',
                                                  variable=self.promiscuous_var,
                                                  onvalue=1, offvalue=0,
                                                  command=self.update_settings,
                                                  state='disabled')
        self.button_promiscuous.grid(column=0, row=4, columnspan=2)

    def run(self) -> None:
        self.root.mainloop()

    def dropdown_opened(self) -> None:
        devices = get_hydralinks()
        self.combobox['values'] = list(devices.keys())

    def update_settings(self) -> None:
        if self.hl is not None:
            speed: Optional[int] = self.speed_var.get()
            if speed not in [100, 1000]:
                speed = None
            master = bool(self.master_var.get())
            promiscuous = bool(self.promiscuous_var.get())

            try:
                self.hl.setup(speed=speed, master=master, promiscuous=promiscuous)
            except Exception as x:
                messagebox.showerror("Error", str(x))

    def device_selected(self) -> None:
        if self.hl is None:
            self.speed_var.set(-1)
            self.master_var.set(-1)
            self.promiscuous_var.set(-1)
            self.button_gigabit['state'] = "disabled"
            self.button_100mbit['state'] = "disabled"
            self.button_master['state'] = "disabled"
            self.button_slave['state'] = "disabled"
            self.button_promiscuous['state'] = "disabled"
        else:
            self.button_gigabit['state'] = "enabled"
            self.button_100mbit['state'] = "enabled"
            self.button_master['state'] = "enabled"
            self.button_slave['state'] = "enabled"
            self.button_promiscuous['state'] = "enabled"

            speed = self.hl.phy.get_speed()
            master = self.hl.phy.get_master()
            promiscuous = (self.hl.mac[0x0b0] & 0x0100) == 0x0100

            if speed is None:
                self.speed_var.set(-1)
            else:
                self.speed_var.set(speed)
            self.master_var.set(1 if master else 0)
            self.promiscuous_var.set(1 if promiscuous else 0)

    def change_device(self, o: tk.Event) -> object:  # type: ignore
        try:
            found = hydralink_by_serial(self.device_var.get())
            self.hl = found
            if found is None:
                messagebox.showerror("Error", "Selected HydraLink was not found")
            self.device_selected()
            return 1
        except Exception as x:
            messagebox.showerror("Error", str(x))


def main() -> None:
    g = Gui()
    g.run()



#  This Source Code Form is subject to the terms of the Mozilla Public
#  License, v. 2.0. If a copy of the MPL was not distributed with this
#  file, You can obtain one at http://mozilla.org/MPL/2.0/.

from ctypes import WinError, sizeof, GetLastError
from ctypes.wintypes import BYTE, DWORD, PDWORD, BOOL, ULONG, PULONG, HANDLE, LPCSTR, LPVOID, PBYTE
import ctypes
import re
from typing import Any, Iterable, List, NamedTuple, Tuple

DIGCF_PRESENT = 2
DIGCF_DEVICEINTERFACE = 16
INVALID_HANDLE_VALUE = 0
SPDRP_DRIVER = 9
ERROR_NO_MORE_ITEMS = 259


class GUID(ctypes.Structure):
    _fields_ = [
        ('Data1', ctypes.c_ulong),
        ('Data2', ctypes.c_ushort),
        ('Data3', ctypes.c_ushort),
        ('Data4', ctypes.c_ubyte*8),
    ]


class SP_DEVICE_INTERFACE_DATA(ctypes.Structure):
    _fields_ = [
        ('cbSize', DWORD),
        ('InterfaceClassGuid', GUID),
        ('Flags', DWORD),
        ('Reserved', PULONG),
    ]


class SP_DEVINFO_DATA(ctypes.Structure):
    _fields_ = [
        ('cbSize', DWORD),
        ('ClassGuid', GUID),
        ('DevInst', DWORD),
        ('Reserved', PULONG)
    ]


class DEVPROPKEY(ctypes.Structure):
    _fields_ = [
        ('fmtid', GUID),
        ('pid', ULONG)
    ]


CreateFile = ctypes.windll.kernel32.CreateFileA
CreateFile.argtypes = [LPCSTR, DWORD, DWORD, LPVOID, DWORD, DWORD, HANDLE]
CreateFile.restype = HANDLE

CloseHandle = ctypes.windll.kernel32.CloseHandle
CloseHandle.argtypes = [HANDLE]
CloseHandle.restype = BOOL

DeviceIoControl = ctypes.windll.kernel32.DeviceIoControl
DeviceIoControl.argtypes = [HANDLE, DWORD, LPVOID, DWORD, LPVOID, DWORD, PDWORD, LPVOID]
DeviceIoControl.restype = BOOL

SetupDiDestroyDeviceInfoList = ctypes.windll.setupapi.SetupDiDestroyDeviceInfoList
SetupDiDestroyDeviceInfoList.argtypes = [LPVOID]
SetupDiDestroyDeviceInfoList.restype = BOOL

SetupDiGetClassDevs = ctypes.windll.setupapi.SetupDiGetClassDevsA
SetupDiGetClassDevs.argtypes = [ctypes.POINTER(GUID), LPCSTR, DWORD, DWORD]
SetupDiGetClassDevs.restype = LPVOID

SetupDiEnumDeviceInterfaces = ctypes.windll.setupapi.SetupDiEnumDeviceInterfaces
SetupDiEnumDeviceInterfaces.argtypes = [LPVOID, ctypes.POINTER(SP_DEVINFO_DATA),
                                        ctypes.POINTER(GUID), DWORD,
                                        ctypes.POINTER(SP_DEVICE_INTERFACE_DATA)]
SetupDiEnumDeviceInterfaces.restype = BOOL

SetupDiGetDeviceInterfaceDetail = ctypes.windll.setupapi.SetupDiGetDeviceInterfaceDetailA
SetupDiGetDeviceInterfaceDetail.argtypes = [LPVOID, ctypes.POINTER(SP_DEVICE_INTERFACE_DATA), LPVOID,
                                            DWORD, PDWORD, ctypes.POINTER(SP_DEVINFO_DATA)]
SetupDiGetDeviceInterfaceDetail.restype = BOOL

SetupDiGetDeviceInterfacePropertyKeys = ctypes.windll.setupapi.SetupDiGetDeviceInterfacePropertyKeys
SetupDiGetDeviceInterfacePropertyKeys.argtypes = [LPVOID, ctypes.POINTER(SP_DEVICE_INTERFACE_DATA),
                                                  ctypes.POINTER(DEVPROPKEY), DWORD, PDWORD, DWORD]
SetupDiGetDeviceInterfacePropertyKeys.restype = BOOL

SetupDiGetDeviceRegistryProperty = ctypes.windll.setupapi.SetupDiGetDeviceRegistryPropertyA
SetupDiGetDeviceRegistryProperty.argtypes = [LPVOID, ctypes.POINTER(SP_DEVINFO_DATA),
                                             DWORD, PDWORD, PBYTE, DWORD, PDWORD]
SetupDiGetDeviceRegistryProperty.restype = BOOL

GUID_DEVINTERFACE_USB_DEVICE = GUID(
    0xa5dcbf10, 0x6530, 0x11d2, (ctypes.c_ubyte*8)(0x90, 0x1f, 0x00, 0xc0, 0x4f, 0xb9, 0x51, 0xed))


def wSetupDiGetDeviceInterfaceDetail(g_hdi: Any, did: SP_DEVICE_INTERFACE_DATA) -> Tuple[bytes, SP_DEVINFO_DATA]:
    dwNeeded = DWORD()
    if not SetupDiGetDeviceInterfaceDetail(
        g_hdi, ctypes.byref(did),
        None, 0, ctypes.byref(dwNeeded),
        None
    ):
        if GetLastError() != 122:
            raise WinError()

    devinfo = SP_DEVINFO_DATA()
    devinfo.cbSize = sizeof(devinfo)
    detailData = (ctypes.c_ubyte*dwNeeded.value)()
    detailDataD = ctypes.cast(detailData, PDWORD)
    detailDataD[0] = 8 if sizeof(PDWORD) == 8 else 6
    devinfo.cbSize = sizeof(devinfo)
    if not SetupDiGetDeviceInterfaceDetail(
        g_hdi, ctypes.byref(did),
        detailData, sizeof(detailData), ctypes.byref(dwNeeded),
        ctypes.byref(devinfo)
    ):
        raise WinError()
    assert dwNeeded.value == sizeof(detailData)
    return bytes(detailData), devinfo


def wSetupDiGetDeviceInterfacePropertyKeys(g_hdi: Any, did: SP_DEVICE_INTERFACE_DATA) -> List[DEVPROPKEY]:
    devpropscount = DWORD()
    if not SetupDiGetDeviceInterfacePropertyKeys(
            g_hdi, ctypes.byref(did), None, 0,
            ctypes.byref(devpropscount), 0):
        if ctypes.GetLastError() != 122:
            raise ctypes.WinError()
    PropertyKeyArray = (DEVPROPKEY * devpropscount.value)()
    if not SetupDiGetDeviceInterfacePropertyKeys(
            g_hdi, ctypes.byref(did), PropertyKeyArray, 4,
            ctypes.byref(devpropscount), 0):
        raise ctypes.WinError()
    assert devpropscount.value == len(PropertyKeyArray)
    return list(PropertyKeyArray)


def wSetupDiGetDeviceRegistryProperty(g_hdi: Any, devinfo: SP_DEVINFO_DATA, property_id: int) -> Tuple[bytes, int]:
    RequiredSize = DWORD()
    PropertyRegDataType = DWORD()
    if not SetupDiGetDeviceRegistryProperty(
            g_hdi, ctypes.byref(devinfo), property_id,
            ctypes.byref(PropertyRegDataType),
            None, 0,
            ctypes.byref(RequiredSize)
    ):
        if ctypes.GetLastError() != 122:
            raise ctypes.WinError()
    PropertyBuffer = (BYTE * RequiredSize.value)()
    if not SetupDiGetDeviceRegistryProperty(
            g_hdi, ctypes.byref(devinfo), property_id,
            ctypes.byref(PropertyRegDataType),
            PropertyBuffer, len(PropertyBuffer),
            ctypes.byref(RequiredSize)
    ):
        raise ctypes.WinError()
    assert RequiredSize.value == len(PropertyBuffer)
    return bytes(PropertyBuffer), PropertyRegDataType.value


def wSetupDiEnumDeviceInterfaces(guid: GUID) -> Iterable[Tuple[Any, SP_DEVICE_INTERFACE_DATA]]:
    guidref = ctypes.byref(guid)
    if (g_hdi := SetupDiGetClassDevs(
        guidref,
        None,
        0,
        DIGCF_DEVICEINTERFACE | DIGCF_PRESENT
    )) == INVALID_HANDLE_VALUE:
        raise WinError()

    dwIndex = 0
    while 1:
        did = SP_DEVICE_INTERFACE_DATA()
        did.cbSize = sizeof(did)

        if not SetupDiEnumDeviceInterfaces(
            g_hdi, None, guidref,
            dwIndex, ctypes.byref(did)
        ):
            if GetLastError() != ERROR_NO_MORE_ITEMS:
                raise WinError()
            break

        yield (g_hdi, did)

        dwIndex += 1

    SetupDiDestroyDeviceInfoList(g_hdi)


class FoundUsbDevice(NamedTuple):
    vid: int
    pid: int
    serialnum: str
    path: str
    software_key: str


def list_usb_devices() -> Iterable[FoundUsbDevice]:
    regex = re.compile(r'^\\\\\?\\usb#vid_([a-f0-9]{4})&pid_([a-f0-9]{4})#([^#]*)' +
                       r'#\{a5dcbf10-6530-11d2-901f-00c04fb951ed\}')
    for g_hdi, did in wSetupDiEnumDeviceInterfaces(GUID_DEVINTERFACE_USB_DEVICE):

        detailData, devinfo = wSetupDiGetDeviceInterfaceDetail(g_hdi, did)
        details = detailData[4:]
        devicePath = details[:details.index(b'\x00')].decode('latin1')

        PropertyBuffer, _ = wSetupDiGetDeviceRegistryProperty(g_hdi, devinfo, SPDRP_DRIVER)

        software_key = PropertyBuffer.decode('latin1')
        software_key = software_key[:software_key.find('\x00')]

        m = regex.match(devicePath)
        assert m is not None
        vid, pid, serialnum = int(m[1], 16), int(m[2], 16), m[3]
        yield FoundUsbDevice(
            vid=vid,
            pid=pid,
            serialnum=serialnum,
            software_key=software_key,
            path=devicePath
        )

    SetupDiDestroyDeviceInfoList(g_hdi)
